# This file was auto-generated by scripts/update_pypi_conda_mapping.py
# DO NOT MANUALLY EDIT
PYPI_TO_CONDA = {
    "Boruta": "boruta_py",
    "Flask-Cors": "flask_cors",
    "Flask-JSON": "flask_json",
    "SoundFile": "pysoundfile",
    "aiohttp_cors": "aiohttp-cors",
    "annoy": "python-annoy",
    "art": "ascii-art",
    "asana": "python-asana",
    "backtrace": "python-backtrace",
    "bash_completion": "py-bash-completion",
    "bilby-pipe": "bilby_pipe",
    "blis": "cython-blis",
    "captest": "pvcaptest",
    "cdo": "python-cdo",
    "click_config_file": "click-config-file",
    "cockroachdb": "cockroachdb-python",
    "conda_lock": "conda-lock",
    "conda_mirror": "conda-mirror",
    "confluent-kafka": "python-confluent-kafka",
    "coreapi": "python-coreapi",
    "coreschema": "python-coreschema",
    "cryptography_vectors": "cryptography-vectors",
    "cufflinks": "python-cufflinks",
    "dask": "dask-core",
    "data-morph-ai": "data-morph-ai",
    "dataclasses": "dataclasses",
    "datadotworld": "datadotworld-py",
    "datalad_container": "datalad-container",
    "delegator.py": "delegator",
    "dials_data": "dials-data",
    "docker": "docker-py",
    "dye_score": "dye-score",
    "eccodes": "python-eccodes",
    "empyrical_dist": "empyrical-dist",
    "enum34": "enum34",
    "esprima": "esprima-python",
    "et_xmlfile": "et-xmlfile",
    "evalml": "evaml-core",
    "exceptiongroup": "exceptiongroup",
    "extract-msg": "msg-extractor",
    "fastjsonschema": "python-fastjsonschema",
    "flair": "python-flair",
    "flake8_nb": "flake8-nb",
    "flatbuffers": "python-flatbuffers",
    "flex": "flex-swagger",
    "flit_core": "flit-core",
    "gnupg": "gnupg-py",
    "google": "googlesearch",
    "google-cloud-bigquery": "google-cloud-bigquery-core",
    "graphviz": "python-graphviz",
    "gssapi": "python-gssapi",
    "h2o": "h2o-py",
    "hdfs": "python-hdfs",
    "hjson": "hjson-py",
    "htcondor": "python-htcondor",
    "iam_units": "iam-units",
    "ib_insync": "ib-insync",
    "import_metadata": "import_metadata",
    "importlib_metadata": "importlib-metadata",
    "installer": "python-installer",
    "intake-astro": "intake-accumulo",
    "intake_geopandas": "intake-geopandas",
    "ioos_tools": "ioos-tools",
    "jsii": "python-jsii",
    "jupyter-client": "jupyter_client",
    "jupyter_jaeger": "jupyter-jaeger",
    "jupyterlab_git": "jupyterlab-git",
    "jupyterlab_latex": "jupyterlab-latex",
    "kaleido": "python-kaleido",
    "kubernetes": "python-kubernetes",
    "libaio": "python-libaio",
    "matplotlib": "matplotlib-base",
    "md_toc": "md-toc",
    "message_ix": "message-ix",
    "msgpack": "msgpack-python",
    "mss": "python-mss",
    "nbconvert_utils": "nbconvert-utils",
    "ndarray_listener": "ndarray-listener",
    "neo": "python-neo",
    "neo4j": "neo4j-python-driver",
    "nest_asyncio": "nest-asyncio",
    "node-semver": "python-node-semver",
    "nvidia-ml-py3": "nvidia-ml",
    "opencv-python": "opencv",
    "ortools": "ortools-python",
    "p-astro": "p_astro",
    "pandas_flavor": "pandas-flavor",
    "paragraph": "python-paragraph",
    "prince": "prince-factor-analysis",
    "pvlib": "pvlib-python",
    "pyct": "pyct-core",
    "pyqt4": "pyqt",
    "pyqt5": "pyqt",
    "pytest_check_links": "pytest-check-links",
    "python-datamatrix": "datamatrix",
    "python-fileinspector": "fileinspector",
    "python-opencv": "opencv",
    "python-pseudorandom": "pseudorandom",
    "python-pygaze": "pygaze",
    "python-qdatamatrix": "qdatamatrix",
    "python-qnotifications": "qnotifications",
    "pywin32": "pywin32-on-windows",
    "quantum-grove": "grove",
    "radio_beam": "radio-beam",
    "redis": "redis-py",
    "rocketpyalpha": "rocketpy",
    "sage_flatsurf": "sage-flatsurf",
    "sagemaker_inference": "sagemaker-inference-toolkit",
    "sagemaker_mxnet_training": "sagemaker_mxnet_container",
    "scikit-build": "scikit-build",
    "scikit-learn": "scikit-learn",
    "setuptools_scm": "setuptools-scm",
    "sounddevice": "python-sounddevice",
    "spherical-functions": "spherical_functions",
    "sphinx-rtd-theme": "sphinx_rtd_theme",
    "stjudecloud-oliver": "oliver",
    "symengine": "python-symengine",
    "tables": "pytables",
    "termstyle": "python-termstyle",
    "torch": "pytorch",
    "torch_geometric": "pytorch_geometric",
    "trino": "trino-python-client",
    "typing": "typing",
    "useDAVE": "dave",
    "vsts": "vsts-python-api",
    "xxhash": "python-xxhash",
}

CONDA_TO_PYPI = {
    "aiohttp-cors": "aiohttp_cors",
    "ascii-art": "art",
    "bilby_pipe": "bilby-pipe",
    "boruta_py": "Boruta",
    "click-config-file": "click_config_file",
    "cockroachdb-python": "cockroachdb",
    "conda-lock": "conda_lock",
    "conda-mirror": "conda_mirror",
    "cryptography-vectors": "cryptography_vectors",
    "cython-blis": "blis",
    "dask-core": "dask",
    "data-morph-ai": "data-morph-ai",
    "dataclasses": "dataclasses",
    "datadotworld-py": "datadotworld",
    "datalad-container": "datalad_container",
    "datamatrix": "python-datamatrix",
    "dave": "useDAVE",
    "delegator": "delegator.py",
    "dials-data": "dials_data",
    "docker-py": "docker",
    "dye-score": "dye_score",
    "empyrical-dist": "empyrical_dist",
    "enum34": "enum34",
    "esprima-python": "esprima",
    "et-xmlfile": "et_xmlfile",
    "evaml-core": "evalml",
    "exceptiongroup": "exceptiongroup",
    "fileinspector": "python-fileinspector",
    "flake8-nb": "flake8_nb",
    "flask_cors": "Flask-Cors",
    "flask_json": "Flask-JSON",
    "flex-swagger": "flex",
    "flit-core": "flit_core",
    "gnupg-py": "gnupg",
    "google-cloud-bigquery-core": "google-cloud-bigquery",
    "googlesearch": "google",
    "grove": "quantum-grove",
    "h2o-py": "h2o",
    "hjson-py": "hjson",
    "iam-units": "iam_units",
    "ib-insync": "ib_insync",
    "import_metadata": "import_metadata",
    "importlib-metadata": "importlib_metadata",
    "intake-accumulo": "intake-astro",
    "intake-geopandas": "intake_geopandas",
    "ioos-tools": "ioos_tools",
    "jupyter-jaeger": "jupyter_jaeger",
    "jupyter_client": "jupyter-client",
    "jupyterlab-git": "jupyterlab_git",
    "jupyterlab-latex": "jupyterlab_latex",
    "matplotlib-base": "matplotlib",
    "md-toc": "md_toc",
    "message-ix": "message_ix",
    "msg-extractor": "extract-msg",
    "msgpack-python": "msgpack",
    "nbconvert-utils": "nbconvert_utils",
    "ndarray-listener": "ndarray_listener",
    "neo4j-python-driver": "neo4j",
    "nest-asyncio": "nest_asyncio",
    "nvidia-ml": "nvidia-ml-py3",
    "oliver": "stjudecloud-oliver",
    "opencv": "python-opencv",
    "ortools-python": "ortools",
    "p_astro": "p-astro",
    "pandas-flavor": "pandas_flavor",
    "prince-factor-analysis": "prince",
    "pseudorandom": "python-pseudorandom",
    "pvcaptest": "captest",
    "pvlib-python": "pvlib",
    "py-bash-completion": "bash_completion",
    "pyct-core": "pyct",
    "pygaze": "python-pygaze",
    "pyqt": "pyqt5",
    "pysoundfile": "SoundFile",
    "pytables": "tables",
    "pytest-check-links": "pytest_check_links",
    "python-annoy": "annoy",
    "python-asana": "asana",
    "python-backtrace": "backtrace",
    "python-cdo": "cdo",
    "python-confluent-kafka": "confluent-kafka",
    "python-coreapi": "coreapi",
    "python-coreschema": "coreschema",
    "python-cufflinks": "cufflinks",
    "python-eccodes": "eccodes",
    "python-fastjsonschema": "fastjsonschema",
    "python-flair": "flair",
    "python-flatbuffers": "flatbuffers",
    "python-graphviz": "graphviz",
    "python-gssapi": "gssapi",
    "python-hdfs": "hdfs",
    "python-htcondor": "htcondor",
    "python-installer": "installer",
    "python-jsii": "jsii",
    "python-kaleido": "kaleido",
    "python-kubernetes": "kubernetes",
    "python-libaio": "libaio",
    "python-mss": "mss",
    "python-neo": "neo",
    "python-node-semver": "node-semver",
    "python-paragraph": "paragraph",
    "python-sounddevice": "sounddevice",
    "python-symengine": "symengine",
    "python-termstyle": "termstyle",
    "python-xxhash": "xxhash",
    "pytorch": "torch",
    "pytorch_geometric": "torch_geometric",
    "pywin32-on-windows": "pywin32",
    "qdatamatrix": "python-qdatamatrix",
    "qnotifications": "python-qnotifications",
    "radio-beam": "radio_beam",
    "redis-py": "redis",
    "rocketpy": "rocketpyalpha",
    "sage-flatsurf": "sage_flatsurf",
    "sagemaker-inference-toolkit": "sagemaker_inference",
    "sagemaker_mxnet_container": "sagemaker_mxnet_training",
    "scikit-build": "scikit-build",
    "scikit-learn": "scikit-learn",
    "setuptools-scm": "setuptools_scm",
    "spherical_functions": "spherical-functions",
    "sphinx_rtd_theme": "sphinx-rtd-theme",
    "trino-python-client": "trino",
    "typing": "typing",
    "vsts-python-api": "vsts",
}
